package sos.stresstest;


import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.text.*;

/**
 * @author Joacim Zschimmer <j@zsch.de>
 * @since 2004-08-10
 * @version 1.0
 * @author Andreas Pschel  <andreas.pueschel@sos-berlin.com>
 * @version 1.1
 * @since 2004-11-13
 *
 * Stresstest ruft eine festgelegte Zeit lang in mehreren Threads Sessions auf.
 * Eine Session ist eine Folge von Aktionen, die eine Sitzung eines Benutzers nachbildet.
 * Stresstest misst die Dauer jeder Aktion und schreibt die Messdaten in eine tabulierte Datei.
 * Beim Aufruf als Hauptprogramm gibt es eine Zusammenfassung auf stderr aus.
 * <p>
 * Fr jeden Thread wird eine Datenbankverbindung geffnet und der Session zur Verfgung gestellt.
 * <p>
 * Jeder Thread ruft nacheinander Sessions auf:
 * <p>
 * <pre>
 *     while( frist noch nicht abgelaufen )
 *     {
 *         StresstestSession session = new Testclass( this );
 *         session.executeChain();
 *     }
 * </pre>
 */

public class Stresstest
{
    //---------------------------------------------------------------------------------------------
    
    final File                  _log_dir;
    final Class                 _test_class;
    final int                   _thread_count;
    final int                   _seconds;
    final boolean               _append_output;
    final File                  _output_file;
    final File                  _configuration_file; 
    final File                  _settings_file; 
    final String				_input_file;
    final String                _param;
    final String                _execute;
    static long                 _start;
    static long		            	_stop;
    static int									_verbosity;
    static long					_current_value;
    static long					_next_value;
    String                      _timestamp; 
    Constructor                 _test_class_constructor;            
    long                        _until_msec;
    StresstestThread[]          _threads;
    boolean                     _write_to_stderr = false;
    
    //-----------------------------------------------------------------------------------get_option
/*    
    public static String get_option( String arg, String option )
    {
        if( arg.startsWith( option ) )
        {
            return arg.substring( option.length() );
        }
        else
            return null;
    }
*/    
    //------------------------------------------------------------------------------string_from_arg
    
/*   public static String string_from_arg( String arg )
    {
        return arg.substring( arg.indexOf( "=" ) + 1 );
    }
    
    //------------------------------------------------------------------------------string_from_arg
    
    public static int int_from_arg( String arg, String option )
    throws
        Exception
    {
        try
        {
            return Integer.parseInt( string_from_arg( arg ) );
        }
        catch( Exception x )
        {
            throw new Exception( "Parameter " + option + " is not a number: " + x, x );
        }
    }
  */  
    class Action_statistics
    {
        String      _action_name;
        int         _count;
        long        _min_msec   = Long.MAX_VALUE;
        long        _max_msec   = 0;
        long        _total_msec = 0;
    }

    //-----------------------------------------------------------------------------------Stresstest
    
    public Stresstest( Class test_class, int thread_count, int seconds, 
                       File configuration_file, File settings_file, File log_dir, File output_file, boolean append_output, String input_file, String param, 
					   String execute, long start, long stop, int verbosity)  
    throws
        Exception
    {
        if( configuration_file    == null )  throw new NullPointerException( "Stresstest: Parameter config == null!" );
        if( settings_file         == null )  throw new NullPointerException( "Stresstest: Parameter settings == null!" );
        if( test_class            == null )  throw new NullPointerException( "Stresstest: Parameter class == null!" );
        if( log_dir               == null )  throw new NullPointerException( "Stresstest: Parameter log-dir == null!" );
        if( output_file           == null )  throw new NullPointerException( "Stresstest: Parameter output == null!" );
        if( input_file            == null )  throw new NullPointerException( "Stresstest: Parameter input == null!" );

        _test_class         = test_class;             
        _thread_count       = thread_count;
        _seconds            = seconds;
        _configuration_file = configuration_file; 
        _settings_file      = settings_file; 
        _log_dir            = log_dir;
        _output_file        = output_file;
        _append_output      = append_output;
        _input_file			= input_file;
        _param				= param;
        _execute			= execute;
        _start 				= start;
        _stop					= stop;
        _verbosity		= verbosity;
        _current_value 		= start;
        _next_value 		= start;

        _test_class_constructor = _test_class.getConstructor( new Class[]{ StresstestThread.class } );        
        _timestamp = new SimpleDateFormat( "yyyy-MM-dd HH:mm:ss" ).format( new Date() );
        
        if (_configuration_file != null && _configuration_file.getName().length() > 0 && !_configuration_file.exists()) {
        	throw new Exception("configuration file not found: " + _configuration_file);
        }
        
        if (_settings_file != null && _settings_file.getName().length() > 0 && !_settings_file.exists()) {
        	throw new Exception("settings file not found: " + _settings_file);
        }

        if ( _test_class.getName().equalsIgnoreCase("sos.stresstest.SOSStresstestSession") && (_input_file == null || _input_file.length() == 0)) {
        	throw new Exception("one of the parameters -class or -input_file must be given");
        }
    }

    //------------------------------------------------------------------------------print_statistic
	/**
	 * Liefert den inkrementierten Wert des Zhlers zurck, dessen Wertebereich mit den Startparametern -start und -stop vereinbart wurde.
	 * Ist der Wert des inkrementierten Zhlers grer als der beim Startparameter -stop angegebene, dann wird der Zhler auf 0 gesetzt 
	 * bzw. auf den Wert des Startparameters -start sofern vorhanden.  
	 * @return inkrementierter Wert des Zhlers oder Startwert.
	 */
	public static long get_next_value() {
		
		if (_stop > 0 && _next_value >= _stop) {
			_next_value = _start;
		} else {
			_next_value++;
		}
		_current_value = _next_value;
		return _next_value;
	}
	
	/**
	 * Liefert den aktuellen Wert des Zhlers zurck, dessen Wertebereich mit den Startparametern -start und -stop vereinbart wurde
	 * @return aktueller Wert des Zhlers.
	 */
	public static long get_current_value() {
		
		return _current_value;
	}

    /**
	 * Liefert den Namen der aktuellen Eingabedatei zurck
	 * @return _current_file Name der Eingabedatei
     */
    public String get_input_file() {
    	return _input_file;
    }
    
    
    /**
	 * Liefert den Wert des Parameters "param" zurck
	 * @return _param Wert des Parameters "param"
     */
    public String get_param() {
    	return _param;
    }
    

    /**
	 * Liefert den Wert des Parameters "execute" zurck
	 * @return _execute Wert des Parameters "execute"
     */
    public String get_execute() {
    	return _execute;
    }
    

    void print_statistic( PrintStream printer )
    {
        Map action_statistics = new HashMap();
        
        for( int i = 0; i < _thread_count; i++ )
        {
            for( Iterator it = _threads[i]._time_records.iterator(); it.hasNext(); )
            {
                StresstestThread.Time_record time_record = (StresstestThread.Time_record)it.next();
                
                Action_statistics stat = (Action_statistics)action_statistics.get( time_record._action_name );
                if( stat == null )
                {
                    stat = new Action_statistics();
                    action_statistics.put( time_record._action_name, stat );
                }

                stat._count++;
                stat._total_msec += time_record._elapsed_msec;
                
                if( stat._min_msec > time_record._elapsed_msec )  stat._min_msec = time_record._elapsed_msec;
                if( stat._max_msec < time_record._elapsed_msec )  stat._max_msec = time_record._elapsed_msec;
            }
        }


        for( Iterator it = action_statistics.entrySet().iterator(); it.hasNext(); )
        {
            Map.Entry         entry       = (Map.Entry)it.next();
            String            action_name = (String)           entry.getKey();
            Action_statistics stat        = (Action_statistics)entry.getValue();
            
            printer.print( action_name );
            printer.print( "  " );
            printer.print( "  calls=" + stat._count );
            printer.print( "  min="   + stat._min_msec  + "ms" );
            printer.print( "  mean="  + ( stat._total_msec / stat._count ) + "ms" );
            printer.print( "  max="   + stat._max_msec  + "ms" );
            printer.print( "\n" );
        }
    }
    
    //------------------------------------------------------------------------------------------run    
    
    public void run() throws Exception
    {
       	System.out.print( "\n");
       	System.out.print( "starting stresstest at " + _timestamp + " ...\n");
       	System.out.print( "using arguments"		
        			+ "  -class=" + _test_class
        			+ "  -config=" + _configuration_file
        			+ "  -settings=" + _settings_file
        			+ "  -log-dir=" + _log_dir 
        			+ "  -output=" + _output_file 
        			+ "  -append-output=" + _append_output 
					+ "  -input=" + _input_file 
					+ "  -param=" + _param 
					+ "  -execute=" + _execute 
					+ "  -threads=" + _thread_count
					+ "  -time=" + _seconds
					+ "  -start=" + _start 
					+ "  -stop=" + _stop
					+ "  -v=" + _verbosity
					+ "\n");
        
        PrintStream output = _output_file.toString().equals( "" )? System.out : new PrintStream( new FileOutputStream( _output_file, _append_output ) );

        _threads = new StresstestThread[ _thread_count ];
        
        for( int i = 0; i < _thread_count; i++ )  _threads[ i ] = new StresstestThread( this, i );
        
        System.out.print( "running ...\n" );
        
        _until_msec = System.currentTimeMillis() + _seconds*1000;
        
        for( int i = 0; i < _thread_count; i++ )  _threads[ i ].start();
        for( int i = 0; i < _thread_count; i++ )  _threads[ i ].join();

        System.out.print( "creating statistics ...\n" );
        print_statistic( System.out );
        
        for( int i = 0; i < _thread_count; i++ )  _threads[ i ].write_data( output );
        
        output.flush();
    }
    
    //-----------------------------------------------------------------------------------------main    
    /** main.
     * 
     * <p>
     * Aufruf:
     * 
     * <p>
     * <b>-class=</b>CLASSNAME
     * <br/>
     * Vollstndiger Name der Testklasse.
     * 
     * <p>
     * <b>-threads=</b>1
     * <br/>
     * Anzahl der gleichzeitig auszufhrenden Threads. Default: 1
     * 
     * <p>
     * <b>-time=</b>60
     * <br/>
     * Zeit in Sekunden, die der Test laufen soll.
     * 
     * <p>
     * <b>-config=</b>config/sos_settings.ini
     * <br/>
     * Pfad der Konfigurationsdatei fr die Datenbankverbindung.
     * 
     * <p>
     * <b>-log-dir=</b>logs
     * <br/>
     * Verzeichnis, in das die Log-Dateien abgelegt werden.
     * 
     * <p>
     * <b>-output=</b>
     * <br/>
     * Pfad der Ausgabedatei (tabuliert) mit den einzelnen Messdaten. Default: stdout
     * 
     * <p>
     * <b>-append-output=</b>true
     * <br/>
     * Anhngen der Ausgaben des Tests in eine vorhandene Ausgabedatei. Default: true
     * 
     * <p>
     * <b>-v=</b>debug-level
     * <br/>
     * Verbosity - Ausfhrliches Logging der Threads mittels debug-level 1-9 einstellen. Default: 0
     * 
     * <p>
     * <b>Beispiel</b>
     * <pre>
     *     java -classpath ... sos.stresstest.Stresstest  -class=sos.stresstest.Test -output=out.csv -threads=20 -time=10 -input=test.sql
     * </pre>
     */
    
    public static void main( String[] args ) throws Exception
    {
        Arguments arguments = new Arguments( args );

        Stresstest stresstest = new Stresstest
        ( 
                Class.forName( arguments.as_string( "-class=",         "sos.stresstest.SOSStresstestSession")), 
                               arguments.as_int   ( "-threads=",       1),
                               arguments.as_int   ( "-time=",          60),
                new File(      arguments.as_string( "-config=",        "") ),
                new File(      arguments.as_string( "-settings=",      "") ),
                new File(      arguments.as_string( "-log-dir=",       "logs/") ),
                new File(      arguments.as_string( "-output=",        "") ),
                               arguments.as_string( "-append-output=", "true").equalsIgnoreCase("true") ? true : false,
                new String(	   arguments.as_string( "-input=",         "")),
                new String(	   arguments.as_string( "-param=",         "")),
                new String(	   arguments.as_string( "-execute=",       "")),
							   							 arguments.as_int   ( "-start=",          0),
							                 arguments.as_int   ( "-stop=",           0),
							                 arguments.as_int   ( "-v=",           		0)	// verbosity 1-9, debug-level fr die Thread-logs
        );

        arguments.check_all_used();

        
        stresstest._write_to_stderr = false;
        stresstest.run();
    }

}
 